# -*- coding: utf-8 -*-
# alone : application framework for small embedded systems.
#               Copyright (c) 2010-2017
#                 Inas Co Ltd., FAR END Technologies Corporation,
#                 All Rights Reserved.
#
# This file is destributed under BSD License. Please read the COPYRIGHT file.
#
# 折れ線グラフと棒グラフに関するクラスを定義
#
# (GraphView) --- (GraphBase) - Graph
#               - Axis -- XAxis
#                      -- YAxis - Y2Axis
#               - LinePlot
#               - BarPlot
#
# DataContainer - ContainerLine
#               - ContainerBar
#

require 'al_graph_base'

module AlGraph

  class AutoScaleErrorX < StandardError; end    # X軸オートスケールエラー
  class AutoScaleErrorY < StandardError; end    # Y軸オートスケールエラー
  class AutoScaleErrorY2 < StandardError; end   # Y2軸オートスケールエラー


  ##
  #  折れ線と棒グラフ用クラス
  #
  class Graph < GraphBase

    # マーカ形状リスト（折れ線グラフのみ）
    SHAPELIST = [:circle, :rectangle, :diamond, :triangle, :cock]

    #@return [Hash]  プロットエリアアトリビュート
    attr_accessor :at_plot_area

    #@return [Hash]  Ｘ軸タイトルアトリビュート
    attr_accessor :at_xaxis_title

    #@return [Hash]  Ｘ軸単位表示アトリビュート
    attr_accessor :at_xaxis_unit

    #@return [Hash]  Ｙ軸タイトルアトリビュート
    attr_accessor :at_yaxis_title

    #@return [Hash]  Ｙ軸単位表示アトリビュート
    attr_accessor :at_yaxis_unit

    #@return [Hash]  Ｙ軸単位表示アトリビュート
    attr_accessor :at_y2axis_unit

    #@return [XAxis]  Ｘ軸オブジェクト
    attr_accessor :x_axis

    #@return [YAxis]  Ｙ軸オブジェクト
    attr_accessor :y_axis

    #@return [Y2Axis]  Ｙ２軸オブジェクト（もしあれば）
    attr_accessor :y2_axis

    #@return [LinePlot]  折れ線グラフオブジェクト
    attr_accessor :line_plot

    #@return [BarPlot]  棒グラフオブジェクト
    attr_accessor :bar_plot

    ##
    # constructor
    # (Graph)
    #
    #@param [Integer] width      幅
    #@param [Integer] height     高さ
    #
    def initialize(width = 320, height = 240)
      super

      @at_plot_area = {:x => 40, :y =>10, :fill => '#eee'}

      # make default
      @at_plot_area[:width] = @width - 50
      @at_plot_area[:width] = 0 if @at_plot_area[:width] < 0
      @at_plot_area[:height] = @height - 30
      @at_plot_area[:Height] = 0 if @at_plot_area[:height] < 0
      @x_axis =
        XAxis.new(@at_plot_area[:width], @at_plot_area[:height], @output)
      @y_axis =
        YAxis.new(@at_plot_area[:width], @at_plot_area[:height], @output)
    end

    ##
    # 折れ線の追加
    # (Graph)
    #
    #@param [Array<Numeric>] ydata   データの配列
    #@param [String] legend          データの名前（凡例）
    #@return [ContainerLine]         データコンテナオブジェクト
    #
    def add_data_line(ydata, legend = nil)
      add_legend() if legend
      @line_plot = LinePlot.new(@output) if ! @line_plot

      data_obj = ContainerLine.new(ydata, legend)
      data_obj.x_axis = @x_axis
      data_obj.y_axis = @y_axis
      data_obj.plot = @line_plot
      data_obj.at_plot_line[:stroke] = data_obj.at_marker[:fill] = @color_list[ @data_series.size % @color_list.size ]
      data_obj.at_marker[:shape] = SHAPELIST[ @data_series.size % SHAPELIST.size ]

      add_data_series(data_obj)
      @line_plot.add_data_series(data_obj)
      @x_axis.add_data_series(data_obj)
      @y_axis.add_data_series(data_obj)

      return data_obj
    end

    ##
    # 折れ線の第２Ｙ軸上への追加
    # (Graph)
    #
    #@param [Array<Numeric>] ydata   データの配列
    #@param [String] legend          データの名前（凡例）
    #@return [ContainerLine]         データコンテナオブジェクト
    #
    def add_data_line_y2(ydata, legend = nil)
      add_legend() if legend
      if ! @y2_axis
        @y2_axis =
          Y2Axis.new(@at_plot_area[:width], @at_plot_area[:height],@output)
        right = @width - @at_plot_area[:width] - @at_plot_area[:x] + 20
        set_margin(nil, right, nil, nil)
      end

      @line_plot = LinePlot.new(@output) if ! @line_plot

      data_obj = ContainerLine.new(ydata, legend)
      data_obj.x_axis = @x_axis
      data_obj.y_axis = @y2_axis
      data_obj.plot = @line_plot
      data_obj.at_plot_line[:stroke] = data_obj.at_marker[:fill] = @color_list[ @data_series.size % @color_list.size ]
      data_obj.at_marker[:shape] = SHAPELIST[ @data_series.size % SHAPELIST.size ]

      add_data_series(data_obj)
      @line_plot.add_data_series(data_obj)
      @x_axis.add_data_series(data_obj)
      @y2_axis.add_data_series(data_obj)

      return data_obj
    end

    ##
    # 棒グラフの追加
    # (Graph)
    #
    #@param [Array<Numeric>] ydata   データの配列
    #@param [String] legend          データの名前（凡例）
    #@param [ContainerBar]           base_bar 積み重ねする場合、ベースになるデータコンテナ
    #@return [ContainerBar]          データコンテナオブジェクト
    #
    def add_data_bar(ydata, legend = nil, base_bar = nil)
      #
      # 積み重ねの場合、Y値を調整。
      #
      if base_bar
        diffsize = base_bar.y_data.size - ydata.size
        ydata.concat( Array.new(diffsize, 0) )  if diffsize > 0

        ydata.each_with_index {|yd, i|
          ydata[i] = 0  if !ydata[i]
          ydata[i] += base_bar.y_data[i]  if base_bar.y_data[i]
        }
      end

      add_legend() if legend

      @bar_plot = BarPlot.new(@output) if ! @bar_plot
      @x_axis.change_mode(:CENTER)

      #
      # コンテナオブジェクトの生成
      #
      data_obj = ContainerBar.new(ydata, legend)
      data_obj.x_axis = @x_axis
      data_obj.y_axis = @y_axis
      data_obj.plot = @bar_plot
      data_obj.at_bar[:fill] = @color_list[ @data_series.size % @color_list.size ]

      #
      # コンテナを配列に保存
      #
      if base_bar
        data_obj.set_stack( base_bar )
        @data_series.each_with_index {|ds, i|
          if ds == base_bar
            @data_series.insert(i, data_obj)
            break
          end
        }
      else
        @data_series << data_obj
      end

      @bar_plot.add_data_series(data_obj, base_bar)
      @x_axis.add_data_series(data_obj)
      @y_axis.add_data_series(data_obj)

      return data_obj
    end

    ##
    # 描画
    # (Graph)
    #
    # 管理下のオブジェクトを次々とcallして、全体を描画する。
    # （ユーザが個々の内部オブジェクトのdrawメソッドを使うことは無い）
    #
    def draw()
      #
      # scaling.
      #
      if !@x_axis.do_scaling
        raise AutoScaleErrorX
      end
      if !@y_axis.do_scaling
        raise AutoScaleErrorY
      end
      if @y2_axis && !@y2_axis.do_scaling
        raise AutoScaleErrorY2
      end

      #
      # draw base items.
      #
      draw_common1()

      #
      # output plot area's clipping path.
      #
      @output.printf(%!<clipPath id="plotarea">\n!)
      @output.printf(%!  <rect x="%d" y="%d" width="%d" height="%d" />\n!, -5, -5, @at_plot_area[:width] + 10, @at_plot_area[:height] + 10)
      @output.printf("</clipPath>\n")

      #
      # grouping in plot items.
      #
      @output.printf(%!<g transform="translate(%d,%d)">\n!, @at_plot_area[:x], @at_plot_area[:y])

      #
      # draw X,Y axis
      #
      @x_axis.draw_z1()
      @y_axis.draw_z1()
      @y2_axis.draw_z1() if @y2_axis

      @x_axis.draw_z2()
      @y_axis.draw_z2()
      @y2_axis.draw_z2() if @y2_axis

      #
      # draw data series.
      #
      @output.printf("\n<!-- draw lines and bars in clipping path -->\n")
      @output.printf(%!<g clip-path="url(#plotarea)">\n!)

      @bar_plot.draw() if @bar_plot
      @line_plot.draw() if @line_plot

      @output.printf("</g><!-- end of clip -->\n")

      #
      # end of group
      #
      @output.printf("</g><!-- end of plot area -->\n\n")


      #
      # draw legend
      #
      if @at_legend
        @output.printf( "\n<!-- draw legends -->\n" )
        if ! @at_legend[:y]
          @at_legend[:y] = (@height - @data_series.size * (@at_legend[:font_size] + @at_legend[:line_spacing])) / 2
          @at_legend[:y] = 0 if @at_legend[:y] < 0
        end

        attr = @at_legend.dup
        attr[:x] += 10
        attr[:y] += attr[:font_size]

        @data_series.each {|ds|
          @output.printf(%!<g%s>\n!, ds.id ? %! id="legend-#{ds.id}"! : "")

          @output.printf("  <text %s>%s</text>\n  ", make_common_attribute_string(attr), Alone::escape_html(ds.legend))
          ds.plot.draw_legend_marker(attr[:x] - 10, (attr[:y] - attr[:font_size] / 3.0).to_i, ds)
          attr[:y] += attr[:font_size] + attr[:line_spacing]
          @output.printf("</g>\n")
        }
        @output.printf("\n")
      end

      #
      # draw x-axis title
      #
      if @at_xaxis_title
        @at_xaxis_title[:x] ||= @at_plot_area[:x] + @at_plot_area[:width] / 2
        @at_xaxis_title[:y] ||= @height - 5

        @output.printf("<text %s>%s</text>\n",
          make_common_attribute_string(@at_xaxis_title),
          Alone::escape_html(@at_xaxis_title[:node_value]))
      end

      #
      # draw x-axis unit
      #
      if @at_xaxis_unit
        @at_xaxis_unit[:x] ||= @at_plot_area[:x] + @at_plot_area[:width]
        @at_xaxis_unit[:y] ||= @height - 5

        @output.printf("<text %s>%s</text>\n",
          make_common_attribute_string(@at_xaxis_unit),
          Alone::escape_html(@at_xaxis_unit[:node_value]))
      end

      #
      # draw y-axis title
      #
      if @at_yaxis_title
        @at_yaxis_title[:x] ||= @at_yaxis_title[:font_size] + 5
        @at_yaxis_title[:y] ||= @at_plot_area[:y] + @at_plot_area[:height] / 2

        @output.printf(%!<text %s transform="rotate(-90,%d,%d)">%s</text>\n!,
          make_common_attribute_string(@at_yaxis_title),
          @at_yaxis_title[:x], @at_yaxis_title[:y],
          Alone::escape_html(@at_yaxis_title[:node_value]))
      end

      #
      # draw y-axis unit
      #
      if @at_yaxis_unit
        @at_yaxis_unit[:x] ||= @at_yaxis_unit[:font_size]
        @at_yaxis_unit[:y] ||= @at_plot_area[:y] - 6

        @output.printf(%!<text %s>%s</text>\n!,
          make_common_attribute_string(@at_yaxis_unit),
          Alone::escape_html(@at_yaxis_unit[:node_value]))
      end

      #
      # draw y2-axis unit
      #
      if @at_y2axis_unit
        @at_y2axis_unit[:x] ||= @at_plot_area[:x] + @at_plot_area[:width]
        @at_y2axis_unit[:y] ||= @at_plot_area[:y] - 6

        @output.printf(%!<text %s>%s</text>\n!,
          make_common_attribute_string(@at_y2axis_unit),
          Alone::escape_html(@at_y2axis_unit[:node_value]))
      end

      draw_common2()
    end

    ##
    # プロットエリアのマージン設定
    # (Graph)
    #
    #@param [Integer] top        上マージン
    #@param [Integer] right      右マージン
    #@param [Integer] bottom     下マージン
    #@param [Integer] left       左マージン
    #
    #  上下左右個別に設定できる。
    #  設定値を変えない場合は、そのパラメータをnilにしてcallする。
    #
    def set_margin(top, right, bottom, left)
      super

      # set axis objects parameters
      @x_axis.width = @at_plot_area[:width]
      @x_axis.height = @at_plot_area[:height]
      @y_axis.width = @at_plot_area[:width]
      @y_axis.height = @at_plot_area[:height]
      if @y2_axis
          @y2_axis.width = @at_plot_area[:width]
          @y2_axis.height = @at_plot_area[:height]
      end
    end

    ##
    # Ｘ軸タイトルの追加
    # (Graph)
    #
    #@param [String] title_string    タイトル文字列
    #
    def add_xaxis_title(title_string)
      set_margin(nil, nil, 35, nil)
      @at_xaxis_title =
        {:node_value => title_string, :font_size => 12, :text_anchor => 'middle'}
    end

    ##
    # Ｘ軸単位表示の追加
    # (Graph)
    #
    #@param [String] unit_string    単位文字列
    #
    def add_xaxis_unit(unit_string)
      set_margin(nil, nil, 35, nil)
      @at_xaxis_unit =
        {:node_value => unit_string, :font_size => 12, :text_anchor => 'middle'}
    end

    ##
    # Ｙ軸タイトルの追加
    # (Graph)
    #
    #@param [String] title_string    タイトル文字列
    #
    def add_yaxis_title(title_string)
      set_margin(nil, nil, nil, 50)
      @at_yaxis_title =
        {:node_value => title_string, :font_size => 12, :text_anchor => 'middle'}
    end

    ##
    # Ｙ軸単位表示の追加
    # (Graph)
    #
    #@param [String] unit_string    単位文字列
    #
    def add_yaxis_unit(unit_string)
      set_margin(25, nil, nil, nil)
      @at_yaxis_unit =
        {:node_value => unit_string, :font_size => 12, :text_anchor => 'start'}
    end


    ##
    # 第２Ｙ軸単位表示の追加
    # (Graph)
    #
    #@param [String] unit_string    単位文字列
    #
    def add_y2axis_unit(unit_string)
      set_margin(25, nil, nil, nil)
      @at_y2axis_unit =
        {:node_value => unit_string, :font_size => 12, :text_anchor => 'start'}
    end

  end



  ##
  #  座標軸スーパークラス
  #
  class Axis < GraphView

    #@return [Hash]  軸アトリビュート
    attr_accessor :at_scale_line

    #@return [Hash]  目盛アトリビュート
    attr_accessor :at_interval_marks

    #@return [Hash]  目盛ラベルアトリビュート
    attr_accessor :at_labels

    #@return [Array]  目盛ラベル
    attr_accessor :labels


    ##
    # constructor
    # (Axis)
    #
    #@param [Integer] width          幅
    #@param [Integer] height         高さ
    #@param [GraphOutput] output     出力制御オブジェクト
    #
    def initialize(width, height, output)
      super
      # 軸アトリビュート
      @at_scale_line = {:stroke => 'black', :stroke_width => 1}
      # 目盛アトリビュート
      @at_interval_marks =
        {:length => 8, :stroke => '#999999', :stroke_width =>1}
      # 目盛ラベルアトリビュート
      @at_labels = {:font_size => 10}
      # 目盛ラベル(配列)
      @labels = []

      @scale_max = nil            # 目盛り最大値
      @scale_min = nil            # 目盛り最小値
      @scale_interval = nil       # 目盛り幅
      @scale_max_user = nil       # ユーザ設定目盛り最大値
      @scale_min_user = nil       # ユーザ設定目盛り最小値
      @scale_interval_user = nil  # ユーザ設定目盛り幅
      @scale_max_min_width= nil   # 最大値-最小値のキャッシュ
      @flag_reverse = false       # 目盛り逆方向（反転）フラグ
    end

    ##
    # 目盛り最大値のユーザ指定
    # (Axis)
    #
    #@param [Numeric] v    目盛り最大値
    #
    def set_max(v)
      @scale_max_user = v
    end
    alias max= set_max

    ##
    # 目盛り最大値を取得する
    # (Axis)
    #
    #@return [Numeric]     現在の目盛り最大値
    #
    def max()
      @scale_max
    end

    ##
    # 目盛り最小値のユーザ指定
    # (Axis)
    #
    #@param [Numeric] v    目盛り最小値
    #
    def set_min(v)
      @scale_min_user = v
    end
    alias min= set_min

    ##
    # 目盛り最小値を取得する
    # (Axis)
    #
    #@return [Numeric]     現在の目盛り最小値
    #
    def min()
      @scale_min
    end

    ##
    # 目盛り幅のユーザ指定
    # (Axis)
    #
    #@param [Numeric] v    目盛り幅
    #
    def set_interval(v)
      @scale_interval_user = v
    end
    alias interval= set_interval

    ##
    # 目盛り逆方向（反転）指示
    # (Axis)
    #
    #@param [Boolean] f    true時、目盛りを反転させる
    #
    def reverse(f = true)
      @flag_reverse = f
    end

    ##
    # グリッド線の付与
    # (Axis)
    #
    def add_grid()
      @at_interval_marks[:grid] = true  if !@at_interval_marks.empty?
    end

    ##
    # グリッド線の消去
    # (Axis)
    #
    def clear_grid()
      @at_interval_marks[:grid] = false  if !@at_interval_marks.empty?
    end

    ##
    # 軸線の消去
    # (Axis)
    #
    def clear_scale_line()
      @at_scale_line.clear
    end

    ##
    # 間隔マークの消去
    # (Axis)
    #
    def clear_interval_marks()
      @at_interval_marks.clear
    end

    ##
    # 目盛ラベルの設定
    # (Axis)
    #
    #@param [Array] labels      目盛ラベル
    #
    def set_labels(labels)
      @labels = labels
    end

    ##
    # 目盛ラベルの消去
    # (Axis)
    #
    def clear_labels()
     @at_labels.clear
    end

    private

    ##
    # 目盛りスケーリング
    # (Axis)
    #
    #@return [Boolean]     成功時、真
    #
    #  サブクラス用 スケーリングサブルーチン
    #
    def do_scaling_sub( min, max )

      # 空のデータ系列が指定されていた場合のmax, min
      max ||= 0
      min ||= 0

      if max == min
        max += max.abs * 0.5
        min -= min.abs * 0.5
      end

      #
      # max point adjustment.
      #
      diff = max - min
      max += diff * 0.1


      #
      # zero point adjustment.
      #
      diff = max - min
      if min > 0 && diff * 2 > min
        min = 0
      elsif (max < 0 && diff * 2 > -max)
        max = 0
      end

      #
      # refrect user settings.
      #
      max = @scale_max_user if @scale_max_user
      min = @scale_min_user if @scale_min_user
      interval ||= @scale_interval_user

      #
      # auto scaling.
      #
      diff = max - min
      if diff > 0
        # calc interval.
        if ! @scale_interval_user
          # intervalがRationalオブジェクトになるverがある？ .to_fで処置
          interval = (10 ** (Math.log10(diff).floor - 1)).to_f
          tick = diff / interval

          if tick > 55
            interval *= 10
          elsif tick > 27
            interval *= 5
          elsif tick > 22
            interval *= 2.5
          elsif tick > 11
            interval *= 2
          end
        end

        #max and min point adjustment.
        max = (max / interval).ceil * interval if ! @scale_max_user
        min = (min / interval).floor * interval if ! @scale_min_user
      end

      return false if ! interval

      #
      # check scale intarval value, and force adjust.
      #
      min_interval = diff.to_f * 5 / @height
      interval = min_interval  if interval < min_interval

      @scale_max = max
      @scale_min = min
      @scale_interval = interval
      @scale_max_min_width = max - min
      @scale_max_min_width = 1 if @scale_max_min_width == 0

      return true
    end

  end


  ##
  # X軸クラス
  #
  class XAxis < Axis

    ##
    # constructor
    # (XAxis)
    #
    #@param [Integer] width          幅
    #@param [Integer] height         高さ
    #@param [GraphOutput] output     出力制御オブジェクト
    #
    def initialize(width, height, output)
      super
      @mode_draw_position = :LEFT
      @at_interval_marks[:grid] = false
    end

    ##
    # 軸描画モード変更
    # (XAxis)
    #
    #@param [Symbol] mode    軸描画モード(:LEFT | :CENTER)
    #
    def change_mode(mode)
      mode_sym = mode.to_sym
      case mode_sym
      when :LEFT, :CENTER
        @mode_draw_position = mode_sym
      end
    end

    ##
    # 目盛り最大値のユーザ指定 (override)
    # (XAxis)
    #
    #@param [Numeric] v   目盛り最大値
    #
    def set_max(v)
      @scale_max_user = v - 1
    end

    ##
    # 目盛り最小値のユーザ指定 (override)
    # (XAxis)
    #
    #@param [Numeric] v   目盛り最小値
    #
    def set_min(v)
      @scale_min_user = v - 1
    end

    ##
    # 目盛りスケーリング
    # (XAxis)
    #
    #@return [Boolean]     成功時、真
    #
    #  あらかじめ与えられているデータ系列情報などを元に、
    #  オートスケール処理など、内部データの整合性をとる。
    #
    def do_scaling()
      @scale_min = 0
      @scale_max = nil
      @scale_interval = 1

      #
      # maximum quantity amang registered datas.
      #

      @data_series.each {|ds|
        num = ds.y_data.size - 1
        @scale_max = num if @scale_max == nil || @scale_max < num
      }
      @scale_max = 1 if @scale_max != nil && @scale_max < 1

      #
      # refrect user settings
      #
      @scale_max = @scale_max_user if @scale_max_user
      @scale_min = @scale_min_user if @scale_min_user
      @scale_interval = @scale_interval_user if @scale_interval_user

      return false if ! @scale_max
      @scale_max_min_width = @scale_max - @scale_min
      @scale_max_min_width = 1 if @scale_max_min_width == 0

      #
      # check scale intarval value, and force adjust.
      #
      if !@scale_interval_user
        min_interval = (@scale_max - @scale_min).to_f * 5 / @width
        min_interval = 1 if min_interval < 1
        @scale_interval = min_interval  if @scale_interval < min_interval
      end

      return true
    end


    ##
    # 軸上のピクセル位置を計算する。
    # (XAxis)
    #
    #@param [Numeric] v    実数
    #@return [Integer]     ピクセル位置
    #
    #  引数が軸上にない場合、返り値も軸上にはないピクセル位置が返る。
    #
    def calc_pixcel_position(v)
      case @mode_draw_position
      when :LEFT
        return (@width * (v - @scale_min) / @scale_max_min_width.to_f).to_i
      when :CENTER
        return ((2 * @width * (v - @scale_min) + @width) / (@scale_max_min_width + 1) / 2.0).to_i
      end
      return 0
    end

    ##
    # 描画　1st pass
    # (XAxis)
    #
    #  スケール描画　パス１。
    #
    #@visibility private
    def draw_z1()
      return if @at_interval_marks.empty?

      #
      # draw interval marks or grid lines
      #
      @output.printf("\n<!-- draw X-axis pass 1 -->\n")
      y1 = @height
      y2 = @height
      if @at_interval_marks[:length] < 0
        y1 = @height - @at_interval_marks[:length]
      else
        y2 = @height - @at_interval_marks[:length]
      end
      if @at_interval_marks[:grid]
        y2 = 0
      end

      @output.printf("<g %s>\n", make_common_attribute_string(@at_interval_marks))

      case @mode_draw_position
      when :LEFT
        i = @scale_min
        while i <= @scale_max do
          x = calc_pixcel_position(i)
          @output.printf(%!  <line x1="%d" y1="%d" x2="%d" y2="%d" />\n!, x, y1, x, y2)
          i += @scale_interval
        end

      when :CENTER
        max_loops = (@scale_max - @scale_min) / @scale_interval + 1
        i = 0
        while i <= max_loops do
          x = (@width * i / max_loops.to_f).to_i
          @output.printf(%!  <line x1="%d" y1="%d" x2="%d" y2="%d" />\n!, x, y1, x, y2)
          i += 1
        end
      end
      @output.printf("</g>\n")
    end

    ##
    # 描画　2nd pass
    # (XAxis)
    #
    #  スケール描画　パス２。
    #
    #@visibility private
    def draw_z2()
      @output.printf("\n<!-- draw X-axis pass 2 -->\n")

      #
      # draw scale line
      #
      if !@at_scale_line.empty?
        @output.printf(%!<line x1="%d" y1="%d" x2="%d" y2="%d" %s/>\n!, 0, @height, @width, @height, make_common_attribute_string(@at_scale_line))
      end

      #
      # draw labels
      #
      if !@at_labels.empty?
        # テキストアンカー位置の決定
        @at_labels[:text_anchor] = 'middle' if ! @at_labels[:text_anchor]

        # 回転させる時は、回転角に応じて自動調整
        if @at_labels[:rotate]
          if 0 < @at_labels[:rotate] && @at_labels[:rotate] < 180
            @at_labels[:text_anchor] = 'start'
          elsif -180 < @at_labels[:rotate] && @at_labels[:rotate] < 0
            @at_labels[:text_anchor] = 'end'
          end
        end

        @output.printf("<g %s>\n", make_common_attribute_string(@at_labels))

        # ラベル出力
        i = -1
        while (v = @scale_min + @scale_interval * (i += 1)) <= @scale_max
          next  if !@labels.empty? && @labels[v] == nil

          x = calc_pixcel_position(v)
          y = @height + @at_labels[:font_size] + 5
          if !@at_labels[:rotate]
            @output.printf('  <text x="%d" y="%d" >', x, y)
          else
            @output.printf('  <text x="%d" y="%d" transform="rotate(%d,%d,%d)" >', x, y, @at_labels[:rotate], x, y - @at_labels[:font_size] / 2)
          end

          if @labels.empty?
            @output.printf( "%d</text>\n", v+1 )
          else
            @output.printf( "%s</text>\n", Alone::escape_html(@labels[v]) )
          end
        end
        @output.printf( "</g>\n" )
      end
    end

  end


  ##
  # Y軸クラス
  #
  class YAxis < Axis

    ##
    # constructor
    # (Yxis)
    #
    #@param [Integer] width        幅
    #@param [Integer] height       高さ
    #@param [GraphOutput] output   出力制御オブジェクト
    #
    def initialize(width, height, output)
      super
      @at_labels[:text_anchor] = 'end'
      @at_interval_marks[:grid] = true
    end

    ##
    # 目盛りスケーリング
    # (YAxis)
    #
    #@return [Boolean]     成功時、真
    #
    #  あらかじめ与えられているデータ系列情報などを元に、
    #  オートスケール処理など、内部データの整合性をとる。
    #
    def do_scaling()
      max = nil
      min = nil

      #
      # search max and min value.
      #
      @data_series.each {|ds|
        next if !ds.y_data_max || !ds.y_data_min
        max = ds.y_data_max if max == nil || max < ds.y_data_max
        min = ds.y_data_min if min == nil || min > ds.y_data_min
      }

      return do_scaling_sub( min, max )
    end

    ##
    # 軸上のピクセル位置を計算する。
    # (YAxis)
    #
    #@param [Numeric] v    実数
    #@return [Integer]     ピクセル位置
    #
    #  引数が軸上にない場合、返り値も軸上にはないピクセル位置が返る。
    #
    def calc_pixcel_position(v)
      y = (@height * (v - @scale_min) / @scale_max_min_width.to_f).to_i
      return @flag_reverse ? y : @height - y
    end

    ##
    # 描画　1st pass
    # (YAxis)
    #
    #  スケール描画　パス１。
    #
    #@visibility private
    def draw_z1()
      return if @at_interval_marks.empty?

      #
      # draw interval marks or grid lines
      #
      @output.printf("\n<!-- draw Y-axis pass 1 -->\n")
      x1 = 0
      x2 = 0
      if @at_interval_marks[:length] < 0
        x1 = @at_interval_marks[:length]
      else
        x2 = @at_interval_marks[:length]
      end
      if @at_interval_marks[:grid]
        x2 = @width
      end

      @output.printf("<g %s>\n", make_common_attribute_string(@at_interval_marks))

      ofs = (@scale_min > 0 || @scale_max < 0) ? @scale_min : 0
      i = 0
      step = 1
      while step >= -1
        v = @scale_interval * i + ofs
        if v > @scale_max || v < @scale_min
          step -= 2
          i = -1
          next
        end

        y = calc_pixcel_position(v)
        @output.printf(%!  <line x1="%d" y1="%d" x2="%d" y2="%d" />\n!, x1, y, x2, y)
        i += step
      end
      @output.printf("</g>\n")
    end

    ##
    # 描画　2nd pass
    # (YAxis)
    #
    #  スケール描画　パス２。
    #
    #@visibility private
    def draw_z2()
      @output.printf("\n<!-- draw Y-axis pass 2 -->\n")

      #
      # draw scale line
      #
      if ! @at_scale_line.empty?
        @output.printf(%!<line x1="%d" y1="%d" x2="%d" y2="%d" %s/>\n!, 0, 0, 0, @height, make_common_attribute_string(@at_scale_line))
      end

      #
      # draw labels
      #
      if !@at_labels.empty?
        @output.printf("<g %s>\n", make_common_attribute_string(@at_labels))

        ofs = (@scale_min > 0 || @scale_max < 0) ? @scale_min : 0
        i = 0
        step = 1
        while step >= -1
          v = @scale_interval * i + ofs
          if v > @scale_max || v < @scale_min
            step -= 2
            i = -1
            next
          end

          @output.printf('  <text x="%d" y="%d">', -5, calc_pixcel_position(v) + @at_labels[:font_size] / 2)

          if !@labels.empty?
            v = @labels[i] ? Alone::escape_html(@labels[i]) : ""
          end

          if @at_labels[:format]
            @output.printf("#{@at_labels[:format]}</text>\n", v)
          else
            @output.printf("%s</text>\n", v.to_s)
          end
          i += step
        end
        @output.printf("</g>\n")
      end
    end

  end


  ##
  # 第２Ｙ軸クラス
  #
  class Y2Axis < YAxis

    ##
    # constructor
    # (Y2xis)
    #
    #@param [Integer] width          幅
    #@param [Integer] height         高さ
    #@param [GraphOutput] output     出力制御オブジェクト
    #
    def initialize(width, height, output)
      super
      @at_labels[:text_anchor] = 'start'
      @at_interval_marks[:grid] = false
    end

    ##
    # 描画　1st pass
    # (Y2Axis)
    #
    #  スケール描画　パス１。
    #
    #@visibility private
    def draw_z1()
      return if @at_interval_marks.empty?

      #
      # draw interval marks or grid lines
      #
      @output.printf("\n<!-- draw Y2-axis pass 1 -->\n")
      x1 = @width
      x2 = @width
      if @at_interval_marks[:length] < 0
        x1 = @width - @at_interval_marks[:length]
      else
        x2 = @width - @at_interval_marks[:length]
      end
      if @at_interval_marks[:grid]
        x2 = 0
      end

      @output.printf("<g %s\n>", make_common_attribute_string(@at_interval_marks))

      ofs = (@scale_min > 0 || @scale_max < 0) ? @scale_min : 0
      i = 0
      step = 1
      while step >= -1
        v = @scale_interval * i + ofs
        if v > @scale_max || v < @scale_min
          step -= 2
          i = -1
          next
        end

        y = calc_pixcel_position(v)
        @output.printf(%!  <line x1="%d" y1="%d" x2="%d" y2="%d" />\n!, x1, y, x2, y)
        i += step
      end
      @output.printf("</g>\n")
    end

    ##
    # 描画　2nd pass
    # (Y2Axis)
    #
    #  スケール描画　パス２。
    #
    #@visibility private
    def draw_z2()
      @output.printf("\n<!-- draw Y2-axis pass 2 -->\n")

      #
      # draw scale line
      #
      if ! @at_scale_line.empty?
        @output.printf(%!<line x1="%d" y1="%d" x2="%d" y2="%d" %s/>\n!, @width, 0, @width, @height, make_common_attribute_string(@at_scale_line))
      end

      #
      # draw labels
      #
      if !@at_labels.empty?
        @output.printf("<g %s>\n", make_common_attribute_string(@at_labels))

        ofs = (@scale_min > 0 || @scale_max < 0) ? @scale_min : 0
        i = 0
        step = 1
        while step >= -1
          v = @scale_interval * i + ofs
          if v > @scale_max || v < @scale_min
            step -= 2
            i = -1
            next
          end

          @output.printf('  <text x="%d" y="%d">', @width + 5, calc_pixcel_position(v) + @at_labels[:font_size] / 2)

          if ! @labels.empty?
            v = @labels[i] ? Alone::escape_html(@labels[i]) : ""
          end

          if @at_labels[:format]
            @output.printf("#{@at_labels[:format]}</text>\n", v)
          else
            @output.printf("%s</text>\n", v.to_s)
          end
          i += step
        end
        @output.printf("</g>\n")
      end
    end
  end



  ##
  #  折れ線グラフプロットクラス
  #
  class LinePlot < GraphView

    ##
    # constructor
    # (LinePlot)
    #
    #@param [GraphOutput] output     出力制御オブジェクト
    #
    def initialize(output)
      super(nil, nil, output)
    end

    ##
    # 描画
    # (LinePlot)
    #
    #  管理下のデータコンテナすべてについて、実際に描画を行う。
    #
    #@visibility private
    def draw()
      @data_series.each {|ds|
        @output.printf("\n<!-- draw line-plot '%s' -->\n", ds.legend)

        #
        # Plot poly line.
        #
        if !ds.at_plot_line.empty?
          @output.printf(%!<polyline%s points="!, ds.id ? %! id="#{ds.id}"! : "")
          ds.y_data.each_with_index {|yd, i|
            next if i < ds.x_axis.min
            break if i > ds.x_axis.max
            next if ! yd

            x = ds.x_axis.calc_pixcel_position(i)
            y = ds.y_axis.calc_pixcel_position(yd)
            @output.printf("%d,%d ", x, y)
          }
          @output.printf(%!" %s/>\n!, make_common_attribute_string(ds.at_plot_line))
        end

        #
        # Plot markers
        #
        if !ds.at_marker.empty?
          @output.printf("<g %s>\n", make_common_attribute_string(ds.at_marker))
          ds.y_data.each_with_index {|yd, i|
            next if i < ds.x_axis.min
            break if i > ds.x_axis.max
            next if ! yd

            x = ds.x_axis.calc_pixcel_position(i)
            y = ds.y_axis.calc_pixcel_position(yd)
            attr = { :shape => ds.at_marker[:shape],
                     :size => ds.at_marker[:size] }
            attr.merge!( ds.at_marker_several[i] ) if ds.at_marker_several[i]
            draw_marker(x, y, attr)
          }
          @output.printf("</g>\n")
        end

        #
        # Data labels.
        #
        if ds.at_data_labels
          @output.printf("<g %s>\n", make_common_attribute_string(ds.at_data_labels))
          ds.y_data.each_with_index {|yd, i|
            next if i < ds.x_axis.min
            break if i > ds.x_axis.max
            next if ! yd

            x = ds.x_axis.calc_pixcel_position(i)
            y = ds.y_axis.calc_pixcel_position(yd)

            case ds.at_data_labels[:position]
            when 'ABOVE'
              y -= 6
            when 'BELOW'
              y += ds.at_data_labels[:font_size] + 6
            when 'LEFT'
              x -= 6
              y += ds.at_data_labels[:font_size] / 2
            when 'RIGHT'
              x += 6
              y += ds.at_data_labels[:font_size] / 2
            when 'CENTER'
              y += ds.at_data_labels[:font_size] / 2
            end

            @output.printf('  <text x="%d" y="%d">', x, y)
            if ds.at_data_labels[:format]
              @output.printf("#{ds.at_data_labels[:format]}</text>", yd)
            else
              @output.printf("%s</text>\n", yd.to_s)
            end
          }

          @output.printf("</g>\n")
        end
      }
    end

    ##
    # マーカーを描画する
    # (LinePlot)
    #
    #@param [Integer] x        Ｘ値
    #@param [Integer] y        Ｙ値
    #@param [Hash] attr        描画アトリビュート
    #@option attr [Symbol] :shape
    #   マーカ種類 (:circle | :rectangle | :diamond | triangle | :cock)
    #@option attr [Integer] :size マーカの大きさ
    #@option attr [String,Integer] :OTHERS その他SVGに出力するアトリビュート
    #
    #@visibility private
    def draw_marker(x, y, attr)
      attrstr = make_common_attribute_string( attr )

      case attr[:shape] && attr[:shape].to_sym
      when :circle
        r = attr[:size] || 4
        @output.printf(%!  <circle cx="%d" cy="%d" r="%d" %s/>\n!, x, y, r, attrstr)

      when :rectangle
        r = attr[:size] || 4
        @output.printf(%!  <rect x="%d" y="%d" width="%d" height="%d" %s/>\n!, x-r, y-r, r*2, r*2, attrstr)

      when :diamond
        r = attr[:size] || 5
        @output.printf(%!  <polygon points="%d,%d %d,%d %d,%d %d,%d" %s/>\n!, x-r, y, x, y-r, x+r, y, x, y+r, attrstr)

      when :triangle
        r = attr[:size] || 5
        @output.printf(%!  <polygon points="%d,%d %d,%d %d,%d" %s/>\n!, x, y-r, x-r, y+r, x+r, y+r, attrstr)

      when :cock
        r = attr[:size] || 4
        @output.printf(%!  <polygon points="%d,%d %d,%d %d,%d %d,%d" %s/>\n!, x-r, y-r, x+r, y+r, x+r, y-r, x-r, y+r, attrstr)

      end
    end

    ##
    # 凡例部マーカー描画
    # (LinePlot)
    #
    #@param [Integer] x    Ｘ値
    #@param [Integer] y    Ｙ値
    #@param [DataContainer] data_obj   データコンテナオブジェクト
    #
    #@visibility private
    def draw_legend_marker(x, y, data_obj)
      if !data_obj.at_plot_line.empty?
        @output.printf(%!<line x1="%d" y1="%d" x2="%d" y2="%d" %s/>\n!, x-9, y, x+9, y, make_common_attribute_string(data_obj.at_plot_line))
      end

      if !data_obj.at_marker.empty?
        attr = { :shape=>data_obj.at_marker[:shape] }
        attr.merge!( data_obj.at_marker )
        draw_marker(x, y, attr)
      end
    end

  end


  ##
  # バーグラフプロットクラス
  #
  class BarPlot < GraphView

    ##
    # constructor
    # (BarPlot)
    #
    #@param [GraphOutput] output   出力制御オブジェクト
    #
    def initialize(output)
      super(nil, nil, output)

      # 棒のオーバーラップ率（％）
      @overlap = 0
      # 棒の間隔率（％：100%で軸とスペースが同じ幅）
      @spacing = 100
    end

    ##
    # データコンテナ追加
    # (BarPlot)
    #
    #@param [ContainerBar] data_obj  データコンテナオブジェクト
    #@param [ContainerBar] base_bar  積み重ねする場合、ベースになるデータコンテナ
    #
    def add_data_series(data_obj, base_bar)
      if base_bar
        @data_series.each_with_index {|ds, i|
          if ds == base_bar
            @data_series.insert(i, data_obj)
            break
          end
        }
      else
        @data_series << data_obj
      end
    end

    ##
    # 棒どうしのオーバーラップ率指定
    # (BarPlot)
    #
    #@param [Integer] v  オーバーラップ率 (%)
    #
    #  0から100を指定する。
    #
    def set_overlap(v)
      @overlap = v
    end
    alias overlap= set_overlap

    ##
    # 棒どうしの間隔指定
    # (BarPlot)
    #
    #@param [Integer] v    間隔率 (%)
    #
    #  0から100を指定する。
    #
    def set_spacing(v)
      @spacing = v
    end
    alias spacing= set_spacing

    ##
    # 描画
    # (BarPlot)
    #
    #  管理下のデータコンテナすべてについて、実際に描画を行う。
    #
    #@visibility private
    def draw()
      num = @data_series.size
      @data_series.each {|ds|
        num -= 1 if ds.base_container
      }

      ov = 1 - @overlap / 100.0   # 棒のオーバーラップ率
      sp = @spacing / 100.0       # 棒幅に対する棒間の率
      w_all = @data_series[0].x_axis.calc_pixcel_position(1) -
        @data_series[0].x_axis.calc_pixcel_position(0) # 全幅 (px)
      w_b = w_all / ( 1 + ov * (num - 1) + sp)  # 棒幅 (px)
      w_s = w_b * sp / 2          # 間隔幅 (px)

      n = 0
      @data_series.each {|ds|
        @output.printf("\n<!-- draw bar-plot '%s' -->\n", ds.legend)

        #
        # Draw bar (1 series)
        #
        @output.printf("<g %s>\n", make_common_attribute_string(ds.at_bar))
        ds.y_data.each_with_index {|yd, i|
          next if i < ds.x_axis.min
          break if i > ds.x_axis.max
          next if ! yd

          x = ds.x_axis.calc_pixcel_position(i) - (w_all / 2.0)
          x1 = x + w_s + n * w_b * ov
          x2 = x1 + w_b
          if ! ds.base_container
            y1 = ds.y_axis.calc_pixcel_position(0)
          else
            y = ds.base_container.y_data[i] || 0
            y1 = ds.y_axis.calc_pixcel_position(y)
          end
          y2 = ds.y_axis.calc_pixcel_position(yd)
          next  if y1 == y2

          @output.printf(%!  <polyline points="%.2f,%.2f %.2f,%.2f %.2f,%.2f %.2f,%.2f"!,
                         x1, y1, x1, y2, x2, y2, x2, y1)
          if ds.at_bar_several[i]
            @output.printf(" %s", make_common_attribute_string(ds.at_bar_several[i]))
          end
          @output.printf(" />\n")
        }
        @output.printf("</g>\n")

        #
        # Data labels.
        #
        if ds.at_data_labels
          @output.printf("<g %s>\n", make_common_attribute_string(ds.at_data_labels))
          ds.y_data.each_with_index {|yd, i|
            next if i < ds.x_axis.min
            break if i > ds.x_axis.max
            next if ! yd

            x = ds.x_axis.calc_pixcel_position(i) - (w_all / 2) + w_s +
              n * w_b * ov
            y = ds.y_axis.calc_pixcel_position(yd)

            case ds.at_data_labels[:position]
            when 'ABOVE'
              x += w_b / 2
              y -= 6
            when 'BELOW'
              x += w_b / 2
              y += dx.at_data_labels[:font_size] + 6
            when 'LEFT'
              x -= 3
              y += ds.at_data_labels[:font_size] / 2
            when 'RIGHT'
              x += w_b + 3
              y += ds.at_data_labels[:font_size] / 2
            when 'CENTER'
              x += w_b / 2
              y += ds.at_data_labels[:font_size] / 2
            end

            @output.printf('  <text x="%d" y="%d">', x, y)
            if ds.at_data_labels[:format]
              @output.printf("#{ds.at_data_labels[:format]}</text>", yd)
            else
              @output.printf("%s</text>\n", yd.to_s)
            end
          }

          @output.printf("</g>\n")
        end

        n += 1 if ! ds.base_container
      }
    end

    ##
    # 凡例部マーカー描画
    # (BarPlot)
    #
    #@param [Integer] x      Ｘ値
    #@param [Integer] y      Ｙ値
    #@param [ContainerBar] data_obj    データコンテナオブジェクト
    #
    #@visibility private
    def draw_legend_marker(x, y, data_obj)
      @output.printf(%!<rect x="%d" y="%d" width="8" height="8" %s/>\n!, x - 3, y - 3, make_common_attribute_string(data_obj.at_bar))
    end
  end



  ##
  #  データコンテナ　スーパークラス
  #
  class DataContainer

    #@return [String] ID
    attr_accessor :id

    #@return [Array<Numeric>] Y値データ
    attr_accessor :y_data

    #@return [Numeric] Y値最大値
    attr_accessor :y_data_max

    #@return [Numeric] Y値最小値
    attr_accessor :y_data_min

    #@return [Hash] データラベルアトリビュート
    attr_accessor :at_data_labels

    #@return [String] 凡例文字列
    attr_accessor :legend

    #@return [XAxis] 使用するＸ軸オブジェクト
    attr_accessor :x_axis

    #@return [YAxis] 使用するＹ軸オブジェクト
    attr_accessor :y_axis

    #@return [LinePlot,BarPlot] 使用するプロットオブジェクト
    attr_accessor :plot


    ##
    # constructor
    # (DataContainer)
    #
    #@param [Array<Numeric>] ydata   Y値データ
    #@param [String] legend          凡例文字列
    #
    def initialize(ydata, legend = nil)
      @at_data_labels = nil

      @y_data = ydata
      (@y_data_min,@y_data_max) = ydata.compact.minmax()
      @legend = legend
    end

    ##
    # 値ラベルを表示
    # (DataContainer)
    #
    #@param [String] pos   値ラベルの位置 (ABOVE | BELOW | LEFT | RIGHT | CENTER)
    #
    #  位置以外は、デフォルト値で表示するよう設定。
    #
    def add_data_labels(pos = 'ABOVE')
      case pos
      when 'ABOVE', 'BELOW', 'CENTER'
        @at_data_labels =
          {:position => pos, :font_size => 9, :text_anchor => 'middle'}
      when 'LEFT'
        @at_data_labels =
          {:position => pos, :font_size => 9, :text_anchor => 'end'}
      when 'RIGHT'
        @at_data_labels =
          {:position => pos, :font_size => 9, :text_anchor => 'start'}
      end
    end
  end


  ##
  # 折れ線グラフ用データコンテナ
  #
  # 線を消してマーカのみの表示にすることもできる。
  #
  class ContainerLine < DataContainer

    #@return [Hash] 線の描画アトリビュート
    attr_accessor :at_plot_line

    #@return [Hash] マーカーの描画アトリビュート
    attr_accessor :at_marker

    #@return [Hash<Hash>] 個別のマーカーの描画アトリビュート
    attr_accessor :at_marker_several

    ##
    # constructor
    # (ContainerLine)
    #
    #@param [Array<Numeric>] ydata   Y値データ
    #@param [String] legend          凡例文字列
    #
    def initialize(ydata, legend = nil)
      super

      @at_plot_line = {:stroke_width=>2, :fill=>:none}
      @at_marker = {:format=>nil, :stroke=>:black, :stroke_width=>2}
      @at_marker_several = {}
    end

    ##
    # 線を表示しない
    # (ContainerLine)
    #
    def clear_line()
      @at_plot_line.clear
    end

    ##
    # マーカーを表示しない
    # (ContainerLine)
    #
    def clear_marker()
      @at_marker.clear
    end

    ##
    # 色の指定
    # (ContainerLine)
    #
    #@param [String] color   色(HTMLカラーコード)
    #
    #  折れ線の場合、線とマーカーの両方の色変えなければならないので、
    #  アトリビュートを2ヶ所変更するよりも簡単にするために作成。
    #
    def set_color(color)
      @at_plot_line[:stroke] = color if !@at_plot_line.empty?
      @at_marker[:fill] = color if !@at_marker.empty?
    end
    alias color= set_color

  end


  ##
  #  バーグラフ用データコンテナ
  #
  class ContainerBar < DataContainer

    #@return [Hash] バーの描画アトリビュート
    attr_accessor :at_bar

    #@return [Hash<Hash>] 個別のバー描画アトリビュート
    attr_accessor :at_bar_several

    #@return [Containerbar] 積み重ねグラフの時、下になるコンテナオブジェクト
    attr_reader :base_container


    ##
    # constructor
    # (ContainerBar)
    #
    #@param [Array<Numeric>] ydata   Y値データ
    #@param [String] legend          凡例文字列
    #
    def initialize(ydata, legend = nil)
      super

      @at_bar = {:stroke_width=>1, :stroke=>:black}
      @at_bar_several = {}
      @base_container = nil
    end

    ##
    # 積み重ね設定
    # (ContainerBar)
    #
    #@param [ContainerBar] base    ベースになるデータコンテナ
    #
    def set_stack(base)
      @base_container = base
    end

    ##
    # 色の指定
    # (ContainerBar)
    #
    #@param [String] color   色(HTMLカラーコード)
    #
    #  ContainerLine::color= との対称性のため定義。
    #
    def set_color(color)
      @at_bar[:fill] = color
    end
    alias color= set_color
  end



  ##
  # AlGraph::Graphのインスタンス生成
  #
  # AlGraph::Graphのインスタンスを生成して返す。
  # AlGraph::Graph.newのかわりにAlGraph.newと書くことができる。
  #
  def self.new(*params)
    AlGraph::Graph.new(*params)
  end

end
