# -*- coding: utf-8 -*-
# alone : application framework for small embedded systems.
#               Copyright (c) 2010-2018
#                 Inas Co Ltd., FAR END Technologies Corporation,
#                 All Rights Reserved.
#
# This file is destributed under BSD License. Please read the COPYRIGHT file.
#
# 散布図クラスを定義
#
# (GraphView) --- (GraphBase) - (Graph) - GraphScatter
#               - (Axis) ----- ScatterXAxis
#               - (LinePlot) - ScatterPlot
#
# (DataContainer) - ContainerScatter
#

require "al_graph"

module AlGraphScatter

  ##
  #  散布図用クラス
  #
  class GraphScatter < AlGraph::Graph

    ##
    # constructor
    # (GraphScatter)
    #
    #@param [Integer] width      幅
    #@param [Integer] height     高さ
    #
    def initialize(width = 320, height = 240)
      super
      @x_axis =
        ScatterXAxis.new(@at_plot_area[:width], @at_plot_area[:height], @output)
    end


    ##
    # データ追加
    # (GraphScatter)
    #
    #@param [Array<Numeric>] data1   データの配列 X軸
    #@param [Array<Numeric>] data2   データの配列 Y軸
    #@param [String] legend          データの名前（凡例）
    #@return [ContainerScatter]      データコンテナオブジェクト
    #
    def add_data(data1, data2, legend = nil)
      add_legend() if legend
      @line_plot = ScatterPlot.new(@output) if ! @line_plot

      data_obj = ContainerScatter.new(data1, data2, legend)
      data_obj.x_axis = @x_axis
      data_obj.y_axis = @y_axis
      data_obj.plot = @line_plot
      data_obj.at_marker[:fill] = @color_list[ @data_series.size % @color_list.size ]
      data_obj.at_marker[:shape] = SHAPELIST[ @data_series.size % SHAPELIST.size ]

      add_data_series(data_obj)
      @line_plot.add_data_series(data_obj)
      @x_axis.add_data_series(data_obj)
      @y_axis.add_data_series(data_obj)

      return data_obj
    end

  end



  ##
  # 散布図用 X軸クラス
  #
  class ScatterXAxis < AlGraph::Axis

    ##
    # constructor
    # (ScatterXAxis)
    #
    #@param [Integer] width          幅
    #@param [Integer] height         高さ
    #@param [GraphOutput] output     出力制御オブジェクト
    #
    def initialize(width, height, output)
      super

      @mode_draw_position = :LEFT
      @at_interval_marks[:grid] = true
    end


    ##
    # 目盛りスケーリング
    # (ScatterXAxis)
    #
    #@return [Boolean]     成功時、真
    #
    #  あらかじめ与えられているデータ系列情報などを元に、
    #  オートスケール処理など、内部データの整合性をとる。
    #
    def do_scaling()
      max = nil
      min = nil

      #
      # search max and min value.
      #
      @data_series.each {|ds|
        next if !ds.x_data_max || !ds.x_data_min
        max = ds.x_data_max if max == nil || max < ds.x_data_max
        min = ds.x_data_min if min == nil || min > ds.x_data_min
      }

      return do_scaling_sub( min, max )
    end


    ##
    # 軸上のピクセル位置を計算する。
    # (ScatterXAxis)
    #
    #@param [Numeric] v    実数
    #@return [Integer]     ピクセル位置
    #
    #  引数が軸上にない場合、返り値も軸上にはないピクセル位置が返る。
    #
    def calc_pixcel_position(v)
      x = (@width * (v - @scale_min) / @scale_max_min_width.to_f).to_i
      return @flag_reverse ? @width - x : x
    end


    ##
    # 描画　1st pass
    # (ScatterXAxis)
    #
    #  スケール描画　パス１。
    #
    #@visibility private
    def draw_z1()
      return if @at_interval_marks.empty?

      #
      # draw interval marks or grid lines
      #
      @output.printf("\n<!-- draw X-axis pass 1 -->\n")
      y1 = @height
      y2 = @height
      if @at_interval_marks[:length] < 0
        y1 = @height - @at_interval_marks[:length]
      else
        y2 = @height - @at_interval_marks[:length]
      end
      if @at_interval_marks[:grid]
        y2 = 0
      end

      @output.printf("<g %s>\n", make_common_attribute_string(@at_interval_marks))

      ofs = (@scale_min > 0 || @scale_max < 0) ? @scale_min : 0
      i = 0
      step = 1
      while step >= -1
        v = @scale_interval * i + ofs
        if v > @scale_max || v < @scale_min
          step -= 2
          i = -1
          next
        end

        x = calc_pixcel_position(v)
        @output.printf(%!  <line x1="%d" y1="%d" x2="%d" y2="%d" />\n!, x, y1, x, y2)
        i += step
      end
      @output.printf("</g>\n")
    end

    ##
    # 描画　2nd pass
    # (ScatterXAxis)
    #
    #  スケール描画　パス２。
    #
    #@visibility private
    def draw_z2()
      @output.printf("\n<!-- draw X-axis pass 2 -->\n")

      #
      # draw scale line
      #
      if !@at_scale_line.empty?
        @output.printf(%!<line x1="%d" y1="%d" x2="%d" y2="%d" %s/>\n!, 0, @height, @width, @height, make_common_attribute_string(@at_scale_line))
      end

      #
      # draw labels
      #
      if !@at_labels.empty?
        # テキストアンカー位置の決定
        @at_labels[:text_anchor] = 'middle' if ! @at_labels[:text_anchor]

        # 回転させる時は、回転角に応じて自動調整
        if @at_labels[:rotate]
          if 0 < @at_labels[:rotate] && @at_labels[:rotate] < 180
            @at_labels[:text_anchor] = 'start'
          elsif -180 < @at_labels[:rotate] && @at_labels[:rotate] < 0
            @at_labels[:text_anchor] = 'end'
          end
        end

        @output.printf("<g %s>\n", make_common_attribute_string(@at_labels))

        ofs = (@scale_min > 0 || @scale_max < 0) ? @scale_min : 0
        i = 0
        step = 1
        while step >= -1
          v = @scale_interval * i + ofs
          if v > @scale_max || v < @scale_min
            step -= 2
            i = -1
            next
          end

          x = calc_pixcel_position(v)
          y = @height + @at_labels[:font_size] + 5
          if !@at_labels[:rotate]
            @output.printf('  <text x="%d" y="%d" >', x, y)
          else
            @output.printf('  <text x="%d" y="%d" transform="rotate(%d,%d,%d)" >', x, y, @at_labels[:rotate], x, y - @at_labels[:font_size] / 2)
          end

          if ! @labels.empty?
            v = @labels[i] ? Alone::escape_html(@labels[i]) : ""
          end

          if @at_labels[:format]
            @output.printf("#{@at_labels[:format]}</text>\n", v)
          else
            @output.printf("%s</text>\n", v.to_s)
          end
          i += step
        end

        @output.printf( "</g>\n" )
      end
    end

  end



  ##
  #  散布図プロットクラス
  #
  class ScatterPlot < AlGraph::LinePlot

    ##
    # 描画
    # (ScatterPlot)
    #
    #  管理下のデータコンテナすべてについて、実際に描画を行う。
    #
    #@visibility private
    def draw()
      @data_series.each {|ds|
        @output.printf("\n<!-- draw Scatter-plot '%s' -->\n", ds.legend)

        #
        # Plot markers
        #
        if !ds.at_marker.empty?
          @output.printf("<g %s>\n", make_common_attribute_string(ds.at_marker))
          ds.y_data.each_with_index {|yd, i|
            xd = ds.x_data[i]
            next if !xd || !yd
            next if xd < ds.x_axis.min
            break if xd > ds.x_axis.max

            x = ds.x_axis.calc_pixcel_position(xd)
            y = ds.y_axis.calc_pixcel_position(yd)
            attr = { :shape => ds.at_marker[:shape],
                     :size => ds.at_marker[:size] }
            attr.merge!( ds.at_marker_several[i] ) if ds.at_marker_several[i]
            draw_marker(x, y, attr)
          }

          @output.printf("</g>\n")
        end
      }
    end

  end



  ##
  # 散布図用データコンテナ
  #
  class ContainerScatter < AlGraph::DataContainer

    #@return [Array<Numeric>] X値データ
    attr_accessor :x_data

    #@return [Numeric] X値最大値
    attr_reader :x_data_max

    #@return [Numeric] X値最小値
    attr_reader :x_data_min

    attr_accessor :at_plot_line  # (互換性用)

    #@return [Hash] マーカーの描画アトリビュート
    attr_accessor :at_marker

    #@return [Hash<Hash>] 個別のマーカーの描画アトリビュート
    attr_accessor :at_marker_several


    ##
    # constructor
    # (ContainerScatter)
    #
    #@param [Array<Numeric>] xdata   X値データ
    #@param [Array<Numeric>] ydata   Y値データ
    #@param [String] legend          凡例文字列
    #
    def initialize(xdata, ydata, legend = nil)
      super(ydata, legend)

      @x_data = xdata
      (@x_data_min,@x_data_max) = xdata.compact.minmax()

      @at_plot_line = {}
      @at_marker = {:format=>nil, :stroke=>:black, :stroke_width=>2}
      @at_marker_several = {}
    end

  end



  ##
  # AlGraphScatter::GraphScatterのインスタンス生成
  #
  # AlGraphScatter::GraphScatterのインスタンスを生成して返す。
  # AlGraphScatter::GraphScatter.newのかわりにAlGraphScatter.newと書くことができる。
  #
  def self.new(*params)
    AlGraphScatter::GraphScatter.new(*params)
  end
end
